///////////////////////////////////////////////////////////////////////////////////////
/// \file migration.cpp
/// \brief contains functions for seed dispersal
///
/// \author Veiko Lehsten
/// $Date: 2018-06-13 19:20:17 +0100  $
///
///////////////////////////////////////////////////////////////////////////////////////
#include <fftw3.h>
#include "guess.h"
#include "migration.h"
#include "guessnc.h"
#include "config.h"		// needed to have min function


fftw_complex *matrix_element_wise_product(fftw_complex * A, fftw_complex * B, size_t size_xy) {
  // standard implementation of Hadamard matrix_element_wise_product
    double scale = 1.0 / size_xy;

    fftw_complex *R = fftw_alloc_complex(size_xy);


    for (int ij = 0; ij < size_xy; ij++) {
        R[ij][0] = (A[ij][0] * B[ij][0] - A[ij][1] * B[ij][1]) * scale;
        R[ij][1] = (A[ij][0] * B[ij][1] + A[ij][1] * B[ij][0]) * scale;
    }

    return R;
}


void generate_seed_kernel_fft(double *kernel, double *kernel_for_conf, size_t, int size_x,
                              int size_y, double short_disp_alpha,
                              double long_disp_alpha,
                              double short_range_disp_frac, int npft,
                              int ipft, double disp_patchsize,
                              bool if_dispersal_fft,
                              bool if_dispersal_float,
                              bool if_dispersal_ext_fft)
{
    int size_xy=size_x*size_y;


    if (short_disp_alpha > 0.0000001 || long_disp_alpha > 0.0000001) {
        // 0 there are no values defined
        double distance;	//distance for seed kernel calculation in m
        //disp_patchsize is in km ^2
        double disp_value;	// tempvalue filled in kernel
        // this here is called for each pft with the seed kernel attributes
        // do the TREEMIG seedkernel first  first without wind
        // start from the corners  text from the working paper:

        //Let L be the matrix containing the number of seeds produced. Another matrix k than needs to have the same dimensions as L.
        //The upper left corner k(1,1) shall contain the proportion of seeds that stay in the gridcell where they are produced, k(1,2)
        //contains the proportion of seeds travelling one grid cell to the south, k(2,1) those that travel one gridcell to the west, k(2,2)
        //those that travel one gridcell southwest and so on. The upper right corner of k contains the proportion of seed travelling one
        //gridcell to the east, the lower right those dispersing to the north east and so on.  Thus the kernel can take different main
        //wind directions into account.

        //  start with upper left corner
        // interpolate linearly between the centers of the gridcells first approach

        //  for (int posy=0; posy<(min((size_y/2),int(200*short_disp_alpha/sqrt(disp_patchsize))))  ;posy++)

        // set to zero to start
        // oder:

        //kernel [0 1 2 3 4 5] = kernel  [0 1 2;
        //				  3 4 5]
	double fine_grid=10;
        for (int posy = 0; posy < size_y; posy++) {
            for (int posx = 0; posx < size_x; posx++) {
                kernel[posx + posy * size_x + ipft * size_xy] = 0.;
            }
        }


        for (int posy = 0; posy < (int (size_y / 2)); posy++) {
            for (int posx = 0; posx < (int (size_x / 2)); posx++) {

	      
	      
	        
                
               if (short_range_disp_frac < 0.999999) {
			disp_value =0.; //initialisation
			for (double fine_x=posx-0.5 ;fine_x<posx+0.5;fine_x+=(1/fine_grid)) {
				for (double fine_y=posy-0.5 ;fine_y<posy+0.5;fine_y+=(1/fine_grid)) {
				  
		
		        	distance = sqrt((fine_x * fine_x) +
                                (fine_y * fine_y)) * sqrt(disp_patchsize * 1000 * 1000);  //take the middle of the cell.
				  
				  disp_value +=
					  (short_range_disp_frac) *
					  (exp(-distance / short_disp_alpha) /
					  (short_disp_alpha * short_disp_alpha)) + ((1 - short_range_disp_frac) *
						  (exp(-distance / long_disp_alpha) /
						    (long_disp_alpha * long_disp_alpha)));
					  
					  
					  
				}
			}
			
		
			
                } else {
			disp_value =0.; //initialisation
			for (double fine_x=posx-0.5 ;fine_x<posx+0.5;fine_x+=(1/fine_grid)) {
				for (double fine_y=posy-0.5 ;fine_y<posy+0.5;fine_y+=(1/fine_grid)) {
					distance = sqrt((fine_x * fine_x) +
					(fine_y * fine_y)) * sqrt(disp_patchsize * 1000 * 1000);  //take the middle of the cell.

					disp_value =
					      (short_range_disp_frac) *
					      (exp(-distance / short_disp_alpha) /
					      (short_disp_alpha * short_disp_alpha));				  
				}
			}	
                }
                
                   
                // fill NW corner
                kernel[posx + posy * size_x + ipft * size_x * size_y] =
                    disp_value;

                if (posy > 0)
                    // fill NE corner
                {
                    kernel[(posx) + (size_y - posy + 1 - 1) * size_x +
                           ipft * size_x * size_y] = disp_value;
                }
                // fill SW corner
                if (posx > 0) {
                    kernel[(size_x - posx) + (posy) * size_x +
                           ipft * size_x * size_y] = disp_value;
                }
                //fill SE corner
                if (posx > 0 && posy > 0) {
                    kernel[(size_x - posx) + (size_y - posy) * size_x +
                           ipft * size_x * size_y] = disp_value;
                }
            }
        }

        //Now standardise all to  sum =1

        double sum_kern = 0.;

        for (int posy = 0; posy < size_y; posy++) {
            for (int posx = 0; posx < size_x; posx++) {
                sum_kern +=
                    kernel[posx + posy * size_x + ipft * size_x * size_y];
            }
        }

        for (int posy = 0; posy < size_y; posy++) {
            for (int posx = 0; posx < size_x; posx++) {
                kernel[posx + posy * size_x +
                       ipft * size_x * size_y] /= sum_kern;
            }
        }

    } else {			// set kernel to zeros
    
        for (int posy = 0; posy < size_y; posy++) {
            for (int posx = 0; posx < size_x; posx++) {
                kernel[posx + posy * size_x + ipft * size_x * size_y] = 0.;
            }
        }
    }
}

double sum(double *arr, size_t count) {
    double _sum = 0;
    for (int i=0; i<count; i++) {
        _sum += arr[i];
    }
    return _sum;
}



void migrate(double *seeds, size_t size, size_t size_x, size_t size_y,
             double *kernel, double probability, int edge_cells,
             bool if_dispersal_fft, bool if_dispersal_float,
             bool if_dispersal_ext_fft, double *terrain,
             int *cell_shift
            )
{
     assert(!((if_dispersal_fft && if_dispersal_float) | (if_dispersal_fft && if_dispersal_ext_fft) || (if_dispersal_float && if_dispersal_ext_fft)));	// max one is allowed to be defined, if non is defined hand the seeds back as feed in the function (don not do anything)


    size_t size_xy = size_x * size_y;

    size_t size_x_ext = size_x + 2*edge_cells;
    size_t size_y_ext = size_y + 2*edge_cells;
    size_t size_xy_ext = (size_y + 2*edge_cells) * (size_x + 2*edge_cells);
    double *seeds_ext = fftw_alloc_real(size_xy_ext);
    // cells also reference themself, hence this should cause no errors
    for (int pos_xy = 0; pos_xy < size_xy; pos_xy++) {
        seeds[pos_xy] = seeds[cell_shift[pos_xy]];
    }
 
    if (if_dispersal_fft) {	// fft
	
	
        // to avoid edge effect both the seed_kernel and the seed_matrix need to be extended by a number of gridcells (edge_cells)
        // due to edge effects, extend the arr by 10 cells in all directions, the kernel is already extended
        double *kernel_ext = fftw_alloc_real(size_xy_ext);

        // copy arr to arr_ext in the middle, leaving the 10 outer posxs and posylums to be zero
        for (int posx = 0; posx < size_x + 2*edge_cells; posx++) {
            for (int posy = 0; posy < size_y + 2*edge_cells; posy++) {
                // set to zero first
                seeds_ext[posx + posy * (size_x + 2*edge_cells)] = 0.;
                kernel_ext[posx + posy * (size_x + 2*edge_cells)] = 0.;
            }
        }

        // copy over
        for (int posx = 0; posx < size_x; posx++) {
            for (int posy = 0; posy < size_y; posy++) {
                seeds_ext[posx + edge_cells + ((posy + edge_cells) * (size_x + 2*edge_cells))] = seeds[posx + posy * size_x];
            }
        }

        // the kernel needs to be copied as well in a way that the corners stay similar to the original kernel

        for (int posx = 0; posx < size_x; posx++) {
            for (int posy = 0; posy < size_y; posy++) {

                int posx_ext = posx;
                if (posx >= size_x / 2) {
                    posx_ext += 2*edge_cells;
                }

                int posy_ext = posy;
                if (posy >= size_y / 2) {
                    posy_ext += 2*edge_cells;
                }
                kernel_ext[posx_ext + posy_ext * (size_x + 2*edge_cells)] = kernel[posx + posy * size_x];
            }
        }

        fftw_complex *out = fftw_alloc_complex(size_xy_ext);
        fftw_complex *seed_out = fftw_alloc_complex(size_xy_ext);

        // the next lines do the following operations
        // new_seeds_to_be_dispersed=ifft2(fft2(seed_produced).*fft2(touruskernel));
        fftw_plan plan = fftw_plan_dft_r2c_2d(size_y_ext, size_x_ext, seeds_ext, out, FFTW_ESTIMATE);
        fftw_execute(plan);

        fftw_plan plan2 = fftw_plan_dft_r2c_2d(size_y_ext, size_x_ext, kernel_ext, seed_out, FFTW_ESTIMATE);
        fftw_execute(plan2);

        fftw_complex *R = matrix_element_wise_product(seed_out, out, size_xy_ext);

        fftw_plan plan3 = fftw_plan_dft_c2r_2d(size_y_ext, size_x_ext, R, seeds_ext, FFTW_ESTIMATE);
        fftw_execute(plan3);

        fftw_free(kernel_ext);
        fftw_free(seed_out);
        fftw_free(out);
        fftw_free(R);
        fftw_destroy_plan(plan);
        fftw_destroy_plan(plan2);
        fftw_destroy_plan(plan3);

        double seeds_before = sum(seeds, size_xy);
        for (int posx = 0; posx < size_x; posx++) {
            for (int posy = 0; posy < size_y; posy++) {
                if (seeds_ext[posx + edge_cells + ((posy + edge_cells) * (size_x + 2*edge_cells))] > 0.01) {
                    seeds[posx + posy * size_x] = seeds_ext[posx + edge_cells + ((posy + edge_cells) * (size_x + 2*edge_cells))];
                } else {
                    seeds[posx + posy * size_x] = 0.;
                }
            }

        }

        fftw_free(seeds_ext);

        double seeds_after = sum(seeds, size_xy);
        if (seeds_after > seeds_before*1.1) {  // allow for some rounding errors
            fail( " fft_migrate more seeds produced than present in the area originally: %g\n",
                  seeds_after / seeds_before);
        }


    } else if (if_dispersal_float)	// floating here the kernel contains not the kernel but the wind field and the kernel data is given as input to this function
    {


        double seed_step_1[size_xy_ext];
        double seed_step_2[size_xy_ext];
        double terrain_ext[size_xy_ext];
        double seed_step_1_remain[size_xy_ext];
        for (int i = 0; i < size_xy_ext; i++) {
            terrain_ext[i] = seed_step_1[i] = seed_step_2[i] = seed_step_1_remain[i]  = 0;	// outside the terrain water there are no seeds)
        }

        // hand seeds to seed_step_1 this is done to increase the area to avoid edge effects over the sea
        // now print it
        for (int posx = 0; posx < size_x; posx++) {
            for (int posy = 0; posy < size_y; posy++) {

                seed_step_1[(posx + edge_cells) +
                            (posy + edge_cells) * size_x_ext] =
                                seeds[posx + posy * size_x];
                seed_step_1_remain[(posx + edge_cells) +
                                   (posy + edge_cells) * size_x_ext] =
                                       seeds[posx + posy * size_x];

                terrain_ext[(posx + edge_cells) +
                            (posy + edge_cells) * size_x_ext] =
                                terrain[posx + posy * size_x];

            }
        }


        int directions_values_x[8]= { 1,  1,  0,  -1,  1,  0, -1, -1 }; // 8 directions
        int directions_values_y[8]= { 1 , 0,  1,   1, -1, -1,  0, -1}; // makes reading easier and finding errors as well

        // from here onwards use always x_large
        for (int steps = 0; steps < 10; steps++) {
            // now place the seeds in the Neuman neighborhood
            for (int posy = 1; posy < size_y_ext - 1; posy++) {	// leave one line at each side
                for (int posx = 1; posx < size_x_ext - 1; posx++) {	// leave one posx at each side
                    for (int directions = 0; directions < 8; directions++) {
                        seed_step_2[posx + posy * size_x_ext]+=seed_step_1[(posx+directions_values_x[directions]) + (posy+directions_values_y[directions]) * size_x_ext] * probability * terrain_ext[posx + posy * size_x_ext];
                        seed_step_1_remain[(posx+directions_values_x[directions]) + (posy+directions_values_y[directions])]-=seed_step_1[(posx+directions_values_x[directions]) + (posy+directions_values_y[directions]) * size_x_ext] * probability * terrain_ext[posx,posy];

                    }


                }
            }


            // now transfer seeds from seed2 to seed1 and set seed 2 to zero

            for (int posy = 0; posy < size_y_ext; posy++) {
                for (int posx = 0; posx < size_x_ext; posx++) {
                    seed_step_1[posx + posy * size_x_ext] =	seed_step_2[posx + posy * size_x_ext]+seed_step_1_remain[posx + posy * size_x_ext];
                    seed_step_1_remain[posx + posy * size_x_ext]=seed_step_1[posx + posy * size_x_ext];
                    seed_step_2[posx + posy * size_x_ext] = 0.;
                }
            }

            // now that we are done cut out the part of the array that contains the actual seeds


        }
        double seeds_before = sum(seeds, size_xy);

        // hand seeds back to seeds from seed_step_1 this is done to increase the area to avoid edge effects over the sea
        for (int posy = 0; posy < size_y; posy++)	// leave one line at each side
        {
            for (int posx = 0; posx < size_x; posx++)	// leave one posx at each side
            {

                seeds[posx + posy * size_x] =
                    seed_step_1[(posx + edge_cells) + (posy + edge_cells) * size_x_ext];

            }
        }

        double seeds_after = sum(seeds, size_xy);
        if (seeds_after > seeds_before*1.1) {  // small proportions are caused by rounding errors  

            // scale all seeds to reach same number
            double scale_factor=seeds_before/seeds_after;
            for (int posy = 0; posy < size_y; posy++)	// leave one line at each side
            {
                for (int posx = 0; posx < size_x; posx++)	// leave one posx at each side
                {

                    seeds[posx + posy * size_x] *=scale_factor;

                }
            }


            seeds_after = sum(seeds, size_xy);


        }
        if (seeds_after > seeds_before*1.1) {  // small proportions are caused by rounding errors
            fail("float:  more seeds produced than present in the area originally: %g %g %g\n",
                 seeds_after / seeds_before, seeds_after , seeds_before );
        }


    }				// end fft versus floating

    else if (if_dispersal_ext_fft) { 

    }


}


void read_terrain( size_t size_x, size_t size_y,double *terrain,  int *cell_shift ) {
    GuessNC::CF::GridcellOrderedVariable* my_var = new GuessNC::CF::GridcellOrderedVariable(param["file_terrain"].str, "terrain");

    int pos_coutner=0;
    for ( int posy=0; posy<size_y ; posy++ ) {
        for ( int posx=0; posx<size_x ; posx++ ) {
            my_var->load_data_for(posx,posy);
            terrain[pos_coutner]=my_var->get_value(0);
            cell_shift[pos_coutner]=my_var->get_value(1);

            //round  to 10 digits make sure zeros stay zeros
            if  (terrain[pos_coutner]<0.0000000001) {
                terrain[pos_coutner]=0.0;
            }
            pos_coutner++;
        }
    }

}
